@@ -32,7 +32,7 @@
 #include <gnutls/crypto.h>
 #include <gnutls_errors.h>
 #include <aes-x86.h>
-#include <x86-common.h>
+#include <x86.h>
 #include <byteswap.h>
 #include <nettle/gcm.h>
 #include <aes-padlock.h>
@@ -69,7 +69,10 @@ static void padlock_aes_set_encrypt_key(struct padlock_ctx *_ctx,
 
 static void aes_gcm_deinit(void *_ctx)
 {
-	gnutls_free(_ctx);
+	struct padlock_ctx *ctx = _ctx;
+
+	zeroize_temp_key(ctx, sizeof(*ctx));
+	gnutls_free(ctx);
 }
 
 static int
